// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

namespace xma {
//////////////////////////////////////////////////////////////////////////////
class ThreadThrowable {
protected:
  ThreadThrowable() {
  }

public:
  virtual ~ThreadThrowable() throw() {
  }

  virtual std::shared_ptr<ThreadThrowable> CreateClone() const = 0;
  virtual void                             Rethrow()           = 0;
};

//////////////////////////////////////////////////////////////////////////////
#define XMA_IMPLEMENT_THREAD_THROWABLE(__TYPE__, __BASE__) \
public: \
  virtual std::shared_ptr<ThreadThrowable> CreateClone() const override { return std::shared_ptr<ThreadThrowable>(new __TYPE__(dynamic_cast<const __TYPE__&>(*this))); } \
  virtual void                             Rethrow() override { throw __TYPE__(dynamic_cast<const __TYPE__&>(*this)); }

//////////////////////////////////////////////////////////////////////////////
class ExceptionStore {
public:
  ExceptionStore() {
  }
  ~ExceptionStore() {
  }

  void SetException(std::shared_ptr<ThreadThrowable> exception) {
    lock_t guard(m_mutex);
    m_exception = exception;
  }
  void RethrowException() {
    lock_t guard(m_mutex);
    if (m_exception) {
      m_exception->Rethrow();
    }
  }
  void Reset() {
    lock_t guard(m_mutex);
    m_exception.reset();
  }
  bool HasException() const {
    lock_t guard(m_mutex);
    return m_exception.get() != nullptr;
  }

private:
  using lock_t = std::lock_guard<std::mutex>;
  mutable std::mutex               m_mutex;
  std::shared_ptr<ThreadThrowable> m_exception;
};

//////////////////////////////////////////////////////////////////////////////
class RuntimeException : public std::runtime_error, public ThreadThrowable {
public:
  XMA_IMPLEMENT_THREAD_THROWABLE(RuntimeException, ThreadThrowable);

public:
  RuntimeException(const std::string& what) : std::runtime_error(what) {
  }
  RuntimeException(const char* what) : std::runtime_error(what) {
  }
  virtual ~RuntimeException() {
  }
};

#define THROW_RUNTIME_EXCEPTION(log, severity, module, ...) \
  { \
    char error_msg[XMA_MAX_LOGMSG_SIZE]; \
    snprintf(error_msg, sizeof(error_msg), __VA_ARGS__); \
    if (log) { \
      log->LogMsg(severity, module, error_msg); \
    } \
    throw xma::RuntimeException(error_msg); \
  }

//////////////////////////////////////////////////////////////////////////////
class BadAllocException : public std::bad_alloc, public ThreadThrowable {
public:
  XMA_IMPLEMENT_THREAD_THROWABLE(BadAllocException, ThreadThrowable);

public:
  BadAllocException() {
  }
  virtual ~BadAllocException() {
  }
};

#define THROW_BAD_ALLOC_EXCEPTION(log, severity, module, ...) \
  { \
    char error_msg[XMA_MAX_LOGMSG_SIZE]; \
    snprintf(error_msg, sizeof(error_msg), __VA_ARGS__); \
    if (log) { \
      log->LogMsg(severity, module, error_msg); \
    } \
    throw BadAllocException(); \
  }

//////////////////////////////////////////////////////////////////////////////
class InvalidArgumentException : public std::invalid_argument, public ThreadThrowable {
public:
  XMA_IMPLEMENT_THREAD_THROWABLE(InvalidArgumentException, ThreadThrowable);

public:
  InvalidArgumentException(const std::string& what) : std::invalid_argument(what) {
  }
  InvalidArgumentException(const char* what) : std::invalid_argument(what) {
  }
  virtual ~InvalidArgumentException() {
  }
};

#define THROW_INVALID_ARGUMENT_EXCEPTION(log, severity, module, ...) \
  { \
    char error_msg[XMA_MAX_LOGMSG_SIZE]; \
    snprintf(error_msg, sizeof(error_msg), __VA_ARGS__); \
    if (log) { \
      log->LogMsg(severity, module, error_msg); \
    } \
    throw InvalidArgumentException(error_msg); \
  }

//////////////////////////////////////////////////////////////////////////////
class OutOfRangeException : public std::out_of_range, public ThreadThrowable {
public:
  XMA_IMPLEMENT_THREAD_THROWABLE(OutOfRangeException, ThreadThrowable);

public:
  OutOfRangeException(const std::string& what) : std::out_of_range(what) {
  }
  OutOfRangeException(const char* what) : std::out_of_range(what) {
  }
  virtual ~OutOfRangeException() {
  }
};

#define THROW_OUT_OF_RANGE_EXCEPTION(log, severity, module, ...) \
  { \
    char error_msg[XMA_MAX_LOGMSG_SIZE]; \
    snprintf(error_msg, sizeof(error_msg), __VA_ARGS__); \
    if (log) { \
      log->LogMsg(severity, module, error_msg); \
    } \
    throw OutOfRangeException(error_msg); \
  }

//////////////////////////////////////////////////////////////////////////////
class LengthErrorException : public std::length_error, public ThreadThrowable {
public:
  XMA_IMPLEMENT_THREAD_THROWABLE(LengthErrorException, ThreadThrowable);

public:
  LengthErrorException(const std::string& what) : std::length_error(what) {
  }
  LengthErrorException(const char* what) : std::length_error(what) {
  }
  virtual ~LengthErrorException() {
  }
};

#define THROW_LENGTH_ERROR_EXCEPTION(log, severity, module, ...) \
  { \
    char error_msg[XMA_MAX_LOGMSG_SIZE]; \
    snprintf(error_msg, sizeof(error_msg), __VA_ARGS__); \
    if (log) { \
      log->LogMsg(severity, module, error_msg); \
    } \
    throw LengthErrorException(error_msg); \
  }

//////////////////////////////////////////////////////////////////////////////
class UnknownException : public ThreadThrowable {
public:
  XMA_IMPLEMENT_THREAD_THROWABLE(UnknownException, ThreadThrowable);

public:
  UnknownException() {
  }
  virtual ~UnknownException() {
  }
};

#define THROW_UNKNOWN_EXCEPTION(log, severity, module, ...) \
  { \
    char error_msg[XMA_MAX_LOGMSG_SIZE]; \
    snprintf(error_msg, sizeof(error_msg), __VA_ARGS__); \
    if (log) { \
      log->LogMsg(severity, module, error_msg); \
    } \
    throw UnknownException(); \
  }

}; // Namespace xma
