// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <cstring>
#include <cstdarg>
#include "xmaapi.h"
#include "xmalogger.h"

#define XMA_MOD "xmaapi"

///////////////////////////////////////////////////////////////////////////////
XmaAPIVersion xma_get_latest_api_version() {
  return XMA_API_VERSION_LATEST;
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_initialize(XmaLogHandle log, XmaInitParameter* init_params, XmaHandle* handle) {
  XmaLogStruct* log_struct = static_cast<XmaLogStruct*>(log);
  TRY()
  log_struct->m_log->LogMsg(XMA_DEBUG_LOG, XMA_MOD, "xma_initialize()");
  XmaSessionFactory* xma_sf = new XmaSessionFactory();
  if (!xma_sf) {
    log_struct->m_log->LogMsg(XMA_DEBUG_LOG, XMA_MOD, "xma_initialize() - Error bad alloc!");
    return XMA_ERROR_BAD_ALLOC;
  }

  xma_sf->m_sf = std::make_shared<xma::InternalSessionFactory>(log_struct->m_log, init_params, true);
  *handle      = static_cast<XmaHandle>(xma_sf);
  log_struct->m_log->LogMsg(XMA_DEBUG_LOG, XMA_MOD, "xma_initialize() - Returning success");
  return XMA_SUCCESS;
  CATCH_ERROR(log_struct->m_log, XMA_MOD)
}

///////////////////////////////////////////////////////////////////////////////
void xma_release(XmaHandle handle) {
  if (!handle) {
    return;
  }

  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(handle);
  xma::LogPtr        log    = xma_sf->m_sf->GetLog();

  TRY()
  log->LogMsg(XMA_DEBUG_LOG, XMA_MOD, "xma_release()");
  delete xma_sf;
  log->LogMsg(XMA_DEBUG_LOG, XMA_MOD, "xma_release() - Returning");
  CATCH(log, XMA_MOD);
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_log_init(XmaLogLevelType log_level, XmaLogType log_type, XmaLogHandle* handle, ...) {
  xma::LogPtr log;
  TRY()
  XmaLogStruct* log_struct = new XmaLogStruct();
  if (!log_struct) {
    return XMA_ERROR_BAD_ALLOC;
  }

  va_list ap;
  va_start(ap, handle);
  log = xma::Log::MakeLogList(log_level, log_type, ap);
  va_end(ap);

  log_struct->m_log = log;
  log->LogMsg(XMA_DEBUG_LOG, XMA_MOD, "xma_log_init()");
  *handle = static_cast<XmaLogHandle>(log_struct);
  log->LogMsg(XMA_DEBUG_LOG, XMA_MOD, "xma_log_init() - Returning");
  return XMA_SUCCESS;
  CATCH_ERROR(log, XMA_MOD);
}

///////////////////////////////////////////////////////////////////////////////
void xma_log_release(XmaLogHandle handle) {
  if (!handle) {
    return;
  }

  XmaLogStruct* log_struct = static_cast<XmaLogStruct*>(handle);
  xma::LogPtr   log        = log_struct->m_log;

  TRY()
  log->LogMsg(XMA_DEBUG_LOG, XMA_MOD, "xma_log_release()");
  delete log_struct;
  log->LogMsg(XMA_DEBUG_LOG, XMA_MOD, "xma_log_release() - Returning");
  CATCH(log, XMA_MOD);
}

///////////////////////////////////////////////////////////////////////////////
void xma_logmsg(XmaLogHandle handle, XmaLogLevelType level, const char* name, const char* msg, ...) {
  XmaLogStruct* log_struct = static_cast<XmaLogStruct*>(handle);
  if (!log_struct) {
    return;
  }

  va_list ap;
  va_start(ap, msg);
  if (log_struct->m_log) {
    log_struct->m_log->LogMsgList(level, name, msg, ap);
  }

  va_end(ap);
}
