// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "xlnx_ml.h"

/**
 * xlnx_ml_cleanup_ml_ctx: Cleanup the ml context - free
 * xma frames, destroy ml session, free xrm resources.
 *
 * @param ctx: The ml context
 */
void xlnx_ml_cleanup_ml_ctx(XlnxMLCtx* ctx)
{
    if (!ctx) {
        return;
    }
    if (ctx->upload_session) {
        xma_filter_session_destroy(ctx->upload_session);
    }
    if (ctx->download_sessions) {
        xma_filter_session_destroy(ctx->download_sessions);
    }
    if (ctx->ml_session) {
        xma_mio_filter_session_destroy(ctx->ml_session);
    }
    if (ctx->handle) {
        xma_release(ctx->handle);
    }
    if (ctx->log) {
        xma_log_release(ctx->log);
    }
    xlnx_ml_cleanup_props(&ctx->ml_xma_props);

    if (ctx->output_xframe_list[0]) {
        xma_frame_free(ctx->output_xframe_list[0]);
    }
    if (ctx->input_xframe[0]) {
        xma_frame_free(ctx->input_xframe[0]);
    }
}

/**
 * xlnx_ml_device_init: Get/allocate xrm resources, xma
 * initialize.
 *
 * @param ctx: The ml context
 * @return XMA_APP_SUCCESS on success XMA_APP_ERROR on error
 */
int32_t xlnx_ml_device_init(XlnxMLCtx* ctx)
{
    int32_t ret = XMA_APP_ERROR;
    XmaInitParameter xma_init_param;
    memset(&xma_init_param, 0, sizeof(XmaInitParameter));
    char m_app_name[32];
    strcpy(m_app_name, XLNX_ML_APP_MODULE);
    xma_init_param.app_name  = m_app_name;
    xma_init_param.device    = ctx->ml_params.device_id;

    XmaParameter params[1];
    uint32_t api_version = XMA_API_VERSION_1_1;

    params[0].name = (char*)XMA_API_VERSION;
    params[0].type = XMA_UINT32;
    params[0].length = sizeof(uint32_t);
    params[0].value = &api_version;

    xma_init_param.params    = params;
    xma_init_param.param_cnt = 1;

    if ((ret = xma_initialize(ctx->log, &xma_init_param, &ctx->handle)) !=
       XMA_SUCCESS) {
        xma_logmsg(ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                   "XMA Initialization failed\n");
        return XMA_APP_ERROR;
    }
    xma_logmsg(ctx->log, XMA_INFO_LOG, XLNX_ML_APP_MODULE,
               "XMA initialization success\n");

    ctx->ml_xma_props.handle = ctx->handle;
    ctx->xma_upload_props.handle = ctx->handle;
    ctx->xma_download_props.handle = ctx->handle;
    
    return ret;
}

/**
 * xlnx_ml_create_frame_props: Allocate the xframes for
 * input/output
 *
 * @param ctx: The ml context
 * @return XMA_APP_SUCCESS on success XMA_APP_ERROR on error.
 */
static int xlnx_ml_create_frame_props(XlnxMLCtx* ctx)
{
    XmaMioFilterProperties props  = ctx->ml_xma_props;
    XmaFrameProperties* fprops = &ctx->input_fprops;
    // Create an input frame for ml
    fprops->width          = props.inputs[0].width;
    fprops->height         = props.inputs[0].height;
    fprops->format         = props.inputs[0].format;
    fprops->sw_format      = props.inputs[0].sw_format;
    for (int plane_id = 0; plane_id < xma_frame_planes_get(ctx->handle, fprops);
        plane_id++) {
        fprops->linesize[plane_id] =
            xma_frame_get_plane_stride(ctx->handle, fprops, plane_id);
    }
    ctx->upload_fprops        = ctx->input_fprops;
    ctx->upload_fprops.format = ctx->upload_fprops.sw_format;

    // Create an output frames for ml
    fprops                 = &ctx->output_fprops;
    fprops->width          = props.inputs[0].width;
    fprops->height         = props.inputs[0].height;
    fprops->format         = props.inputs[0].format;
    fprops->sw_format      = props.inputs[0].sw_format;

    for (int plane_id = 0;
         plane_id < xma_frame_planes_get(ctx->handle, fprops); plane_id++) {
        fprops->linesize[plane_id] =
            xma_frame_get_plane_stride(ctx->handle, fprops, plane_id);
    }

    ctx->download_fprops = ctx->output_fprops;
    ctx->download_fprops.format = ctx->download_fprops.sw_format;
    
    return XMA_APP_SUCCESS;
}

/**
 * xlnx_ml_create_ml_ctx: Uses ctx->ml_params to create the 
 * rest of the ml context. Does not init the device/create 
 * session. 
 * @param ctx: A pointer to the ml context
 * @return XMA_APP_SUCCESS on success XMA_APP_ERROR on error
 */
int32_t xlnx_ml_create_ml_ctx(XlnxMLCtx* ctx)
{
    int ret = XMA_APP_ERROR;
  
    ret = xlnx_ml_create_xma_props(
        ctx->handle, &ctx->ml_params, &ctx->xma_upload_props,
        &ctx->xma_download_props, &ctx->ml_xma_props);
    if (ret != XMA_APP_SUCCESS) {
        return ret;
    }
    ret = xlnx_ml_create_frame_props(ctx);
    return ret;
}

/**
 * xlnx_ml_create_ml_sessions: Creates the ml sessions 
 * using the ml props.
 * @param ctx: A pointer to the ml context
 * @return XMA_APP_SUCCESS on success XMA_APP_ERROR on error
 */
int32_t xlnx_ml_create_ml_sessions(XlnxMLCtx* ctx)
{
    ctx->xma_upload_props.handle = ctx->handle;
    ctx->upload_session = xma_filter_session_create(&ctx->xma_upload_props);
    if (!ctx->upload_session) {
        xma_logmsg(ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                   "Failed to create upload session\n");
        return XMA_APP_ERROR;
    }

    ctx->xma_download_props.handle = ctx->handle;
    ctx->download_sessions =
        xma_filter_session_create(&ctx->xma_download_props);
    if (!ctx->download_sessions) {
        xma_logmsg(ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                   "Failed to create download session\n");
        return XMA_APP_ERROR;
    }
    
    // Create an ml session based on the requested properties
    ctx->ml_xma_props.handle = ctx->handle; 
    ctx->ml_session =
        xma_mio_filter_session_create(&ctx->ml_xma_props);
    if (!ctx->ml_session) {
        xma_logmsg(ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                   "Failed to create ml session\n");
        return XMA_APP_ERROR;
    }
    
    return XMA_APP_SUCCESS;
}
